(* Consider the function gen1 below. *)

let gen1 (l:int list) : int -> int =
  fun (x:int) -> List.fold_left (fun acc next -> (next * x) + acc) 0 l

(* Now consider the following client code that uses the function gen1: *)

let data =
  [
    [1;2;3;4;5;6;7;8;9;0;1;2;3;4;5;6;7;8;9;0];
    [0;9;8;7;6;5;4;3;2;1];
    [];
    [1;3;2;4;6;7];
    [1;2;3;4;5;6;7;8;9;0]
  ]

let output = List.map gen1 data

(*
Write a new function gen2 that when used by the client code in place
of gen1 results in a substantially smaller (in terms of space) value
output.  Your new function gen2 must have the same type as gen1 and
must be functionally equivalent to gen1 (but you don't have to prove
that).

First, in a sentence or two, explain your main idea for reducing the
size of output - even if you can't quite figure out how to rewrite the
code correctly, you can explain a good way one might try to save
space.  Also, if you need to make any assumptions about your solution
state them here clearly.

Explanation:

  First, recognize that output is a list of functions. For each
  function, the space cost is the cost of its closure, that is: its
  definition (body) and its environment.

  Looking at the anonymous function of x, analyze what the free
  variables are: List.fold_left (we've written fold_left! It has no free
  variables, so it costs insignificant space), and l (potentially huge
  amount of space if the list is long).

  So the main space cost is the cost of the list l that winds up in each
  of the functions' closures.  Hence, the best techniques for optimizing the
  space cost of representing these functions is to find a way to avoid
  including the list in the function closure.

  As it turns out, all gen1 does is return a function that, if given an
  int argument x, will produce the sum of the list elements after they
  are scaled by x. So you can avoid including the list in the function
  closure by computing the sum of the elements in the list first, and
  then multiplying the sum by x in the function.

Write a test to check that the output of gen2 matches the output of gen1.
*)

let gen2 (l:int list) : int -> int =
 (* pre-compute the sum of the elements of the list *)
 let n = List.fold_left (fun acc next -> next + acc) 0 l in
 (* now, create closure that includes one int, not a big list *)
 (fun (x:int) ->  x * n)

(* you can figure out how to write a test on your own! *)
